#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v3 transfer action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class AcceptTransferRequest(command.ShowOne):
    _description = _("Accept volume transfer request.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer_request',
            metavar="<transfer-request-id>",
            help=_('Volume transfer request to accept (ID only)'),
        )
        parser.add_argument(
            '--auth-key',
            metavar="<key>",
            required=True,
            help=_('Volume transfer request authentication key'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        try:
            transfer_request_id = volume_client.find_transfer(
                parsed_args.transfer_request, ignore_missing=False
            ).id
        except sdk_exceptions.ResourceNotFound:
            # Non-admin users will fail to lookup name -> ID so we just
            # move on and attempt with the user-supplied information
            transfer_request_id = parsed_args.transfer_request

        result = volume_client.accept_transfer(
            transfer_request_id, parsed_args.auth_key
        )
        columns = ('id', 'name', 'volume_id')
        data = (result.id, result.name, result.volume_id)
        return columns, data


class CreateTransferRequest(command.ShowOne):
    _description = _("Create volume transfer request.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar="<name>",
            help=_('New transfer request name (default to None)'),
        )
        parser.add_argument(
            '--snapshots',
            action='store_true',
            dest='snapshots',
            help=_(
                'Allow transfer volumes without snapshots (default) '
                '(supported by --os-volume-api-version 3.55 or later)'
            ),
            default=None,
        )
        parser.add_argument(
            '--no-snapshots',
            action='store_false',
            dest='snapshots',
            help=_(
                'Disallow transfer volumes without snapshots '
                '(supported by --os-volume-api-version 3.55 or later)'
            ),
        )
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_('Volume to transfer (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        attrs: dict[str, Any] = {}

        if parsed_args.snapshots is not None:
            if not sdk_utils.supports_microversion(volume_client, '3.55'):
                msg = _(
                    "--os-volume-api-version 3.55 or greater is required to "
                    "support the '--(no-)snapshots' option"
                )
                raise exceptions.CommandError(msg)

            # unfortunately this option is negative so we have to reverse
            # things
            attrs['no_snapshots'] = not parsed_args.snapshots

        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )
        result = volume_client.create_transfer(
            volume_id=volume.id,
            name=parsed_args.name,
            **attrs,
        )
        columns = ('auth_key', 'created_at', 'id', 'name', 'volume_id')
        data = (
            result.auth_key,
            result.created_at,
            result.id,
            result.name,
            result.volume_id,
        )
        return columns, data


class DeleteTransferRequest(command.Command):
    _description = _("Delete volume transfer request(s).")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer_request',
            metavar="<transfer-request>",
            nargs="+",
            help=_('Volume transfer request(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = 0

        for t in parsed_args.transfer_request:
            try:
                volume_client.delete_transfer(t, ignore_missing=False)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete volume transfer request "
                        "with name or ID '%(transfer)s': %(e)s"
                    ),
                    {'transfer': t, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.transfer_request)
            msg = _(
                "%(result)s of %(total)s volume transfer requests failed"
                " to delete"
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListTransferRequest(command.Lister):
    _description = _("Lists all volume transfer requests.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            dest='all_projects',
            action="store_true",
            default=False,
            help=_('Include all projects (admin only)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        column_headers = ('ID', 'Name', 'Volume')
        columns = ('id', 'name', 'volume_id')

        results = volume_client.transfers(
            details=True,
            all_projects=parsed_args.all_projects,
        )
        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in results),
        )


class ShowTransferRequest(command.ShowOne):
    _description = _("Show volume transfer request details.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'transfer_request',
            metavar="<transfer-request>",
            help=_('Volume transfer request to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = volume_client.find_transfer(
            parsed_args.transfer_request, ignore_missing=False
        )
        columns = ('created_at', 'id', 'name', 'volume_id')
        data = (result.created_at, result.id, result.name, result.volume_id)
        return columns, data
